/*
Created by: X
http://www.createdbyx.com/

This work is licensed under a Creative Commons Attribution-Noncommercial-Share Alike 2.5 Canada License 
http://creativecommons.org/licenses/by-nc-sa/2.5/ca/. 
Permissions beyond the scope of this license may be available at http://contact.createdbyx.com/.
*/
using System.Collections;
using System.Collections.Generic;
using System;
using System.ComponentModel;


namespace xCollectiveXNA
{
    /// <summary>
    /// A generic index based collection that implements the IList interface.
    /// </summary>
    /// <typeparam name="T">The type that will be stored in the collection.</typeparam>
    /// <remarks>This collection is not thread safe. It creates and reuses a reference to a ValueEventArgs 
    /// object in order to reduce garbage collection.</remarks>
    public class GenericCollection<T> : IList<T>, IDisposable
    {
        /// <summary>
        /// This event is raised before a item is added to the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> BeforeAdded;
        /// <summary>
        /// This event is raised after a item has been added to the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> AfterAdded;
        /// <summary>
        /// This event is raised before a item is inserted into the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> BeforeInsert;
        /// <summary>
        /// This event is raised after a item has been inserted into the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> AfterInsert;
        /// <summary>
        /// This event is raised before a item is removed from the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> BeforeRemove;
        /// <summary>
        /// This event is raised after a item has been removed from the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> AfterRemove;
        /// <summary>
        /// This event is raised before a item is removed from the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> BeforeRemoveAt;
        /// <summary>
        /// This event is raised after a item has been removed from the collection.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> AfterRemoveAt;
        /// <summary>
        /// This event is raised before a item in the collection is changed.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> BeforeChange;
        /// <summary>
        /// This event is raised after an item in the collection is changed.
        /// </summary>
        public event EventHandler<ValueEventArgs<T>> AfterChange;
        /// <summary>
        /// This event is raised before the collection is cleared.
        /// </summary>
        public event EventHandler<CancelEventArgs> BeforeClear;
        /// <summary>
        /// This event is raised after the collection is cleared.
        /// </summary>
        public event EventHandler AfterClear;

        /// <summary>
        /// Used to hold the items in the collection.
        /// </summary>
        protected internal List<T> listItems;
        /// <summary>
        /// Used when raising events.
        /// </summary>
        protected internal ValueEventArgs<T> args = new ValueEventArgs<T>();
        /// <summary>
        /// Used when raising clear events.
        /// </summary>
        protected internal CancelEventArgs cancelArgs = new CancelEventArgs();


        /// <summary>
        /// Gets or sets the list that will be used to store and retrieve items.
        /// </summary>
        public List<T> List
        {
            get { return this.listItems; }
            set { this.listItems = value; }
        }

        ///<summary>
        /// Initializes a new instance of the GenericCollection class.
        ///</summary>
        public GenericCollection()
        {
            this.listItems = new List<T>();
        }

        ///<summary>
        /// Initializes a new instance of the GenericCollection class.
        ///</summary>
        ///<param name="list">A reference to a list object that will be used to store and retrieve items.</param>
        /// <remarks>Providing your own list allows the generic collection to act as a wrapper to a existing list.</remarks>
        public GenericCollection(List<T> list)
            : this()
        {
            this.listItems = list;
        }

        /// <summary>
        /// Sorts the elements in the entire collection using the default comparer.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">The default comparer System.Collections.Generic.Comparer<T>.Default cannot find an implementation of the System.IComparable<T> generic interface or the System.IComparable interface for type T.</exception>
        public void Sort()
        {
            this.listItems.Sort();
        }

        /// <summary>
        /// Sorts the elements in the entire System.Collections.Generic.List<T> using the specified System.Comparison<T>.
        /// </summary>
        /// <param name="comparison">The System.Comparison<T> to use when comparing elements.</param>
        public void Sort(Comparison<T> comparison)
        {
            this.listItems.Sort(comparison);
        }

        /// <summary>
        /// Sorts the elements in the entire System.Collections.Generic.List<T> using the specified comparer.
        /// </summary>
        /// <param name="comparer"> The System.Collections.Generic.IComparer<T> implementation to use when comparing elements, or null to use the default comparer System.Collections.Generic.Comparer<T>.Default.</param>
        public void Sort(IComparer<T> comparer)
        {
            this.listItems.Sort(comparer);
        }

        /// <summary>
        /// Sorts the elements in a range of elements in System.Collections.Generic.List<T> using the specified comparer.
        /// </summary>
        /// <param name="index">The zero-based starting index of the range to sort.</param>
        /// <param name="count">The length of the range to sort.</param>
        /// <param name="comparer">The System.Collections.Generic.IComparer<T> implementation to use when comparing elements, or null to use the default comparer System.Collections.Generic.Comparer<T>.Default.</param>
        public void Sort(int index, int count, IComparer<T> comparer)
        {
            this.listItems.Sort(index, count, comparer);
        }

        /// <summary>
        /// Adds an item to the collection.
        /// </summary>
        /// <param name="item">The item to be added.</param>
        public virtual void Add(T item)
        {
            if (!this.DoBeforeEvent(this.AfterAdded, item, -1)) return;

            // add item
            this.listItems.Add(item);

            this.DoAfterEvent(this.AfterAdded, item, -1);
        }

        /// <summary>
        /// Used internally to raise "before" events.
        /// </summary>
        /// <param name="eventHandler">A reference to an event handler.</param>
        /// <param name="item">A reference to a item.</param>
        /// <param name="index">A index in the collection.</param>
        /// <returns>true if all the delagates in the invocation list did not cancel otherwise false.</returns>
        protected internal bool DoBeforeEvent(EventHandler<ValueEventArgs<T>> eventHandler, T item, int index)
        {
            // check if any delegates are attached
            if (eventHandler != null)
            {
                // invoke each attached delegate or until canceled
                foreach (var handler in (EventHandler<ValueEventArgs<T>>[])eventHandler.GetInvocationList())
                {
                    // setup args
                    this.args.Cancel = false;
                    this.args.Index = index;
                    this.args.Value = item;

                    // raise event
                    handler(this, this.args);

                    // check if canceled
                    if (args.Cancel) return false;
                }
            }
            return true;
        }

        /// <summary>
        /// Used internally to raise "after" events.
        /// </summary>
        /// <param name="eventHandler">A reference to an event handler.</param>
        /// <param name="item">A reference to a item.</param>
        /// <param name="index">A index in the collection.</param>
        /// <returns>true if all the delagates in the invocation list did not cancel otherwise false.</returns>
        protected internal void DoAfterEvent(EventHandler<ValueEventArgs<T>> eventHandler, T item, int index)
        {
            // raise event
            if (eventHandler != null)
            {
                // invoke each attached delegate  
                foreach (var handler in (EventHandler<ValueEventArgs<T>>[])eventHandler.GetInvocationList())
                {
                    // setup args
                    this.args.Cancel = false;
                    this.args.Index = index;
                    this.args.Value = item;

                    // raise event
                    handler(this, this.args);
                }
            }
        }

        ///<summary>
        /// Adds an item to the collection.
        ///</summary>
        ///<param name="item">The item to add to the collection.</param>
        public virtual void Add(T[] item)
        {
            foreach (var tmp in item)
                this.Add(tmp);
        }

        ///<summary>
        /// Copies the items from the collection into an array.
        ///</summary>
        ///<returns>Returns the items in the collection in the form of an array.</returns>
        public T[] ToArray()
        {
            return this.listItems.ToArray();
        }

        /// <summary>
        /// Gets the number of items in the collection.
        /// </summary>
        public virtual int Count
        {
            get { return this.listItems.Count; }
        }

        ///<summary>
        /// Gets a value indicating whether the collection is read-only.
        ///</summary>
        ///<returns>
        ///true if the is read-only; otherwise, false.
        ///</returns>
        public bool IsReadOnly
        {
            get { return false; }
        }

        ///<summary>
        /// Returns an enumerator that iterates through the collection.
        ///</summary>
        ///<returns>
        /// A IEnumerator that can be used to iterate through the collection.
        ///</returns>
        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            return this.listItems.GetEnumerator();
        }

        ///<summary>
        /// Returns an enumerator that iterates through a collection.
        ///</summary>
        ///<returns>
        /// An <see cref="T:System.Collections.IEnumerator" /> object that can be used to iterate through the collection.
        ///</returns>
        public virtual IEnumerator GetEnumerator()
        {
            return this.listItems.GetEnumerator();
        }

        ///<summary>
        /// Removes all items from the collection.
        ///</summary>
        ///<exception cref="T:System.NotSupportedException">
        ///  The collection is read-only. 
        ///  </exception>
        public virtual void Clear()
        {
            // check if any delegates are attached
            if (this.BeforeClear != null)
            {
                // invoke each attached delegate or until canceled
                foreach (var handler in (EventHandler<CancelEventArgs>[])this.BeforeClear.GetInvocationList())
                {
                    // setup args
                    this.cancelArgs.Cancel = false;

                    // raise event
                    handler(this, this.cancelArgs);

                    // check if canceled
                    if (this.cancelArgs.Cancel) return;
                }
            }

            // clear items
            this.listItems.Clear();

            // raise event
            if (this.AfterClear != null)
            {
                // invoke each attached delegate  
                foreach (var handler in (EventHandler[])this.AfterClear.GetInvocationList())
                {
                    // setup args
                    this.args.Cancel = false;

                    // raise event
                    handler(this, this.args);
                }
            }
        }

        ///<summary>
        /// Removes the first occurrence of a specific object from the <see cref="T:System.Collections.Generic.ICollection`1" />.
        ///</summary>
        ///<returns>
        /// true if <paramref name="item" /> was successfully removed from the <see cref="T:System.Collections.Generic.ICollection`1" />; otherwise, false. This method also returns false if <paramref name="item" /> is not found in the original <see cref="T:System.Collections.Generic.ICollection`1" />.
        ///</returns>
        ///<param name="item">The object to remove from the <see cref="T:System.Collections.Generic.ICollection`1" />.</param>
        ///<exception cref="T:System.NotSupportedException">The <see cref="T:System.Collections.Generic.ICollection`1" /> is read-only.</exception>
        public virtual bool Remove(T item)
        {
            int index = this.listItems.IndexOf(item);
            if (index == -1) return false;

            if (!this.DoBeforeEvent(this.BeforeRemove, item, index)) return false;

            // remove item
            this.listItems.Remove(item);

            this.DoAfterEvent(this.AfterRemove, item, index);
            return true;
        }


        ///<summary>
        /// Determines the index of a specific item in the <see cref="T:System.Collections.Generic.IList`1" />.
        ///</summary>
        ///<returns>
        /// The index of <paramref name="item" /> if found in the list; otherwise, -1.
        ///</returns>
        ///<param name="item">The object to locate in the <see cref="T:System.Collections.Generic.IList`1" />.</param>
        public virtual int IndexOf(T item)
        {
            return this.listItems.IndexOf(item);
        }

        ///<summary>
        /// Inserts an item to the <see cref="T:System.Collections.Generic.IList`1" /> at the specified index.
        ///</summary>
        ///<param name="index">The zero-based index at which <paramref name="item" /> should be inserted.</param>
        ///<param name="item">The object to insert into the <see cref="T:System.Collections.Generic.IList`1" />.</param>
        ///<exception cref="T:System.ArgumentOutOfRangeException">
        /// <paramref name="index" /> is not a valid index in the <see cref="T:System.Collections.Generic.IList`1" />.
        ///</exception>
        ///<exception cref="T:System.NotSupportedException">The <see cref="T:System.Collections.Generic.IList`1" /> is read-only.</exception>
        public virtual void Insert(int index, T item)
        {
            if (!this.DoBeforeEvent(this.BeforeInsert, item, index)) return;

            // insert item
            this.listItems.Insert(index, item);

            this.DoAfterEvent(this.AfterInsert, item, index);
        }

        ///<summary>
        /// Gets or sets the element at the specified index.
        ///</summary>
        ///<returns>
        /// The element at the specified index.
        ///</returns>
        ///<param name="index">The zero-based index of the element to get or set.</param>
        ///<exception cref="T:System.ArgumentOutOfRangeException"><paramref name="index" /> is not a valid index in the <see cref="T:System.Collections.Generic.IList`1" />.
        /// </exception>
        ///<exception cref="T:System.NotSupportedException">
        /// The property is set and the <see cref="T:System.Collections.Generic.IList`1" /> is read-only.
        /// </exception>
        public virtual T this[int index]
        {
            get { return this.listItems[index]; }
            set
            {
                if (!this.DoBeforeEvent(this.BeforeChange, value, index)) return;

                // set item
                this.listItems[index] = value;

                // raise after event
                this.DoAfterEvent(this.AfterChange, value, index);
            }
        }

        /// <summary>
        /// Removes a item from the collection.
        /// </summary>
        /// <param name="index">The index of the item to remove.</param>
        public virtual void RemoveAt(int index)
        {
            T itemRef = this.listItems[index];

            if (!this.DoBeforeEvent(this.BeforeRemoveAt, itemRef, index)) return;

            // remove item
            this.listItems.RemoveAt(index);

            this.DoBeforeEvent(this.AfterRemoveAt, itemRef, index);
        }

        ///<summary>
        /// Copies the elements of the collection to an array, starting at a particular array index.
        ///</summary>
        ///<param name="items">The one-dimensional array that is the destination of the elements copied from the collection. The array must have zero-based indexing.</param>
        ///<param name="arrayindex">The zero-based index in items at which copying begins. </param>
        ///<exception cref="T:System.ArgumentNullException">items is null. </exception>
        ///<exception cref="T:System.ArgumentOutOfRangeException">arrayindex is less than 0.  </exception>
        ///<exception cref="T:System.ArgumentException">items is multidimensional.
        ///                
        ///                    -or-
        ///                    arrayindex is equal to or greater than the length of items.
        ///                
        ///                    -or-
        ///                
        ///                    The number of elements in the source collection is greater than the available space from arrayIndex to the end of the destination array.
        ///                
        ///                    -or-
        ///                
        ///                    Cannot be cast automatically to the type of the destination items.
        ///                </exception>
        public virtual void CopyTo(T[] items, int arrayindex)
        {
            this.listItems.CopyTo(items, arrayindex);
        }

        ///<summary>
        /// Determines whether the collection contains a specific value.
        ///</summary>
        ///<param name="item">The object to locate in the collection.</param>
        ///<returns>
        ///true if item is found in the collection otherwise, false.
        ///</returns>
        public virtual bool Contains(T item)
        {
            return this.listItems.Contains(item);
        }


        #region " IDisposable Support "
        private bool disposedValue; // To detect redundant calls

        // IDisposable
        protected virtual void Dispose(bool disposing)
        {
            if (!this.disposedValue)
            {
                if (disposing)
                {
                    // TODO: free unmanaged resources when explicitly called
                    this.listItems = null;
                }

                // TODO: free shared unmanaged resources
            }
            this.disposedValue = true;
        }

        // This code added by Visual Basic to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion
    }

}
