/*
Created by: X
http://www.createdbyx.com/

xCollectiveXNA is licenced under the Creative Commons Attribution-NonCommercial-ShareAlike 2.5 Canada
http://creativecommons.org/licenses/by-nc-sa/2.5/ca/
*/
using System.Collections;
using System.Collections.Generic;
using System;
using System.ComponentModel;

namespace xCollectiveXNA
{
    /// <summary>
    /// A generic dictionary class that supports events.
    /// </summary>
    /// <typeparam name="TKey">Specifies a key type.</typeparam>
    /// <typeparam name="TValue">Specifies the value type.</typeparam>
    public class GenericDictionary<TKey, TValue> : IDictionary<TKey, TValue>, IDisposable
    {

        /// <summary>
        /// This event is raised before a item is added to the dictionary.
        /// </summary>
        public event EventHandler<KeyValueEventArgs<TKey, TValue>> BeforeAdded;
        /// <summary>
        /// This event is raised after a item is added to the dictionary.
        /// </summary>
        public event EventHandler<KeyValueEventArgs<TKey, TValue>> AfterAdded;
        /// <summary>
        /// This event is raised before a item is removed to the dictionary.
        /// </summary>
        public event EventHandler<KeyValueEventArgs<TKey, TValue>> BeforeRemove;
        /// <summary>
        /// This event is raised after a item is removed to the dictionary.
        /// </summary>
        public event EventHandler<KeyValueEventArgs<TKey, TValue>> AfterRemove;
        /// <summary>
        /// This event is raised before a item in the dictionary is changed.
        /// </summary>
        public event EventHandler<KeyValueEventArgs<TKey, TValue>> BeforeChange;
        /// <summary>
        /// This event is raised after a item in the dictionary was changed.
        /// </summary>
        public event EventHandler<KeyValueEventArgs<TKey, TValue>> AfterChange;
        /// <summary>
        /// This event is raised before al item in the dictionary are removed.
        /// </summary>
        public event EventHandler<CancelEventArgs> BeforeClear;
        /// <summary>
        /// This event is raised after all items in the dictionary were removed.
        /// </summary>
        public event EventHandler AfterClear;
        /// <summary>
        /// This event is raised before a key in the dictionary is changed.
        /// </summary>
        public event EventHandler<NewKeyValueEventArgs<TKey, TValue>> BeforeKeyChanged;
        /// <summary>
        /// This event is raised after a key in the dictionary was changed.
        /// </summary>
        public event EventHandler<NewKeyValueEventArgs<TKey, TValue>> AfterKeyChanged;

        /// <summary>
        /// used to hold items in the dictionary
        /// </summary>
        protected internal Dictionary<TKey, TValue> dictionaryItems;
        /// <summary>
        /// Used when raising events.
        /// </summary>
        protected internal KeyValueEventArgs<TKey, TValue> args = new KeyValueEventArgs<TKey, TValue>();
        /// <summary>
        /// Used when raising events.
        /// </summary>
        protected internal NewKeyValueEventArgs<TKey, TValue> newArgs = new NewKeyValueEventArgs<TKey, TValue>();
        /// <summary>
        /// Used when raising events.
        /// </summary>
        protected internal CancelEventArgs cancelArgs = new CancelEventArgs();


        /// <summary>
        /// Gets or sets the dictionary that will be used to store and retrieve items.
        /// </summary>
        public Dictionary<TKey, TValue> Dictionary
        {
            get { return this.dictionaryItems; }
            set { this.dictionaryItems = value; }
        }

        ///<summary>
        /// Initializes a new instance of the GenericDictionary class.
        ///</summary>
        public GenericDictionary()
        {
            this.dictionaryItems = new Dictionary<TKey, TValue>();
        }

        ///<summary>
        /// Initializes a new instance of the GenericDictionary class.
        ///</summary>
        ///<param name="dictionary">A reference to a dictionary object that will be used to store and retrieve items.</param>
        /// <remarks>Providing your own dictionary allows the generic dictionary to act as a wrapper to a existing dictionary.</remarks>
        public GenericDictionary(Dictionary<TKey, TValue> dictionary)
            : this()
        {
            this.dictionaryItems = dictionary;
        }

        /// <summary>
        /// Used internally to raise "before" events.
        /// </summary>
        /// <param name="eventHandler">A reference to an event handler.</param>
        /// <param name="key">A key in the dictionary.</param>
        /// <param name="item">A reference to a item.</param>
        /// <returns>true if all the delagates in the invocation list did not cancel otherwise false.</returns>
        protected internal bool DoBeforeEvent(EventHandler<KeyValueEventArgs<TKey, TValue>> eventHandler, TKey key, TValue item)
        {
            // check if any delegates are attached
            if (eventHandler != null)
            {
                // invoke each attached delegate or until canceled
                foreach (var handler in (EventHandler<KeyValueEventArgs<TKey, TValue>>[])eventHandler.GetInvocationList())
                {
                    // setup args
                    this.args.Cancel = false;
                    this.args.Key = key;
                    this.args.Value = item;

                    // raise event
                    handler(this, this.args);

                    // check if canceled
                    if (args.Cancel) return false;
                }
            }
            return true;
        }

        /// <summary>
        /// Used internally to raise "before" events.
        /// </summary>
        /// <param name="eventHandler">A reference to an event handler.</param>
        /// <param name="key">A existing key in the dictionary.</param>
        /// <param name="newkey">A new key to be added to the dictionary.</param>
        /// <param name="item">A reference to a item.</param>
        /// <returns>true if all the delagates in the invocation list did not cancel otherwise false.</returns>
        protected internal bool DoBeforeEvent(EventHandler<NewKeyValueEventArgs<TKey, TValue>> eventHandler, TKey key, TKey newkey, TValue item)
        {
            // check if any delegates are attached
            if (eventHandler != null)
            {
                // invoke each attached delegate or until canceled
                foreach (var handler in (EventHandler<NewKeyValueEventArgs<TKey, TValue>>[])eventHandler.GetInvocationList())
                {
                    // setup args
                    this.newArgs.Cancel = false;
                    this.newArgs.Key = key;
                    this.newArgs.Value = item;

                    // raise event
                    handler(this, this.newArgs);

                    // check if canceled
                    if (args.Cancel) return false;
                }
            }
            return true;
        }

        /// <summary>
        /// Used internally to raise "after" events.
        /// </summary>
        /// <param name="eventHandler">A reference to an event handler.</param>
        /// <param name="item">A reference to a item.</param>
        /// <param name="key">A key in the dictionary.</param>
        /// <returns>true if all the delagates in the invocation list did not cancel otherwise false.</returns>
        protected internal void DoAfterEvent(EventHandler<KeyValueEventArgs<TKey, TValue>> eventHandler, TValue item, TKey key)
        {
            // raise event
            if (eventHandler != null)
            {
                // invoke each attached delegate  
                foreach (var handler in (EventHandler<KeyValueEventArgs<TKey, TValue>>[])eventHandler.GetInvocationList())
                {
                    // setup args
                    this.args.Cancel = false;
                    this.args.Key = key;
                    this.args.Value = item;

                    // raise event
                    handler(this, this.args);
                }
            }
        }

        /// <summary>
        /// Used internally to raise "after" events.
        /// </summary>
        /// <param name="eventHandler">A reference to an event handler.</param>
        /// <param name="item">A reference to a item.</param>
        /// <param name="key">A existing key in the dictionary.</param>
        /// <param name="newkey">A new key in the dictionary.</param>
        /// <returns>true if all the delagates in the invocation list did not cancel otherwise false.</returns>
        protected internal void DoAfterEvent(EventHandler<NewKeyValueEventArgs<TKey, TValue>> eventHandler, TValue item, TKey key, TKey newkey)
        {
            // raise event
            if (eventHandler != null)
            {
                // invoke each attached delegate  
                foreach (var handler in (EventHandler<NewKeyValueEventArgs<TKey, TValue>>[])eventHandler.GetInvocationList())
                {
                    // setup args
                    this.newArgs.Cancel = false;
                    this.newArgs.Key = key;
                    this.newArgs.Value = item;

                    // raise event
                    handler(this, this.newArgs);
                }
            }
        }

        /// <summary>
        /// Renames a existing key to a new one.
        /// </summary>
        /// <param name="existingkey">The existing key specifier.</param>
        /// <param name="newkey">The value that the new key will be set to.</param>
        /// <returns></returns>
        public virtual bool ChangeKey(TKey existingkey, TKey newkey)
        {
            // if existing key is not present then exit
            if (!this.ContainsKey(existingkey)) throw new KeyNotFoundException("The existingkey parameter does not exist in the dictionary.");
            // if new key is already present then exit
            if (this.ContainsKey(newkey)) throw new ArgumentException("An element with the same key already exists in the dictionary.");

            // get existing item
            TValue item = this.dictionaryItems[existingkey];

            // raise before event
            if (!this.DoBeforeEvent(this.BeforeKeyChanged, existingkey, newkey, item)) return false;

            // remove existing item
            this.dictionaryItems.Remove(existingkey);
            // add existing item with new key
            this.dictionaryItems.Add(newkey, item);

            // raise after KeyChanged event 
            this.DoAfterEvent(this.AfterKeyChanged, item, existingkey, newkey);

            return true;
        }


        ///<summary>
        /// Gets a KeyCollection of all the keys in the dictionary.
        ///</summary>
        public Dictionary<TKey, TValue>.KeyCollection Keys
        {
            get { return this.dictionaryItems.Keys; }
        }

        /// <summary>
        /// Gets a ValueCollection of all the items in the dictionary.
        /// </summary>
        public Dictionary<TKey, TValue>.ValueCollection Values
        {
            get { return this.dictionaryItems.Values; }
        }

        ///<summary>
        ///Adds an element with the provided key and value to the <see cref="T:System.Collections.Generic.IDictionary`2" />.
        ///</summary>
        ///<param name="Key">The object to use as the key of the element to add.</param>
        ///<param name="item">The object to use as the value of the element to add.</param>
        ///<exception cref="T:System.ArgumentNullException">
        /// key is null.
        ///  </exception>
        ///<exception cref="T:System.ArgumentException">
        /// An element with the same key already exists in the dictionary." />.
        /// </exception>
        ///<exception cref="T:System.NotSupportedException">
        ///  The dictionary is read-only.
        ///  </exception>
        public virtual void Add(TKey Key, TValue item)
        {
            if (!this.DoBeforeEvent(this.BeforeAdded, Key, item)) return;

            // add item
            this.dictionaryItems.Add(Key, item);

            this.DoAfterEvent(this.AfterAdded, item, Key);
        }


        ///<summary>
        ///  Adds an item to the dictionary.
        ///</summary>
        ///<param name="pair">The object to add to the dictionary.</param>
        ///<exception cref="T:System.NotSupportedException">The dictionary is read-only.</exception>
        public virtual void Add(KeyValuePair<TKey, TValue> pair)
        {
            this.Add(pair.Key, pair.Value);
        }

        ///<summary>
        ///Removes all items from the dictionary.
        ///</summary>
        ///<exception cref="T:System.NotSupportedException">
        ///  The dictionary is read-only. 
        /// </exception>
        public virtual void Clear()
        {
            // check if any delegates are attached
            if (this.BeforeClear != null)
            {
                // invoke each attached delegate or until canceled
                foreach (var handler in (EventHandler<CancelEventArgs>[])this.BeforeClear.GetInvocationList())
                {
                    // setup args
                    this.cancelArgs.Cancel = false;

                    // raise event
                    handler(this, this.cancelArgs);

                    // check if canceled
                    if (this.cancelArgs.Cancel) return;
                }
            }

            // clear the dictionary
            this.dictionaryItems.Clear();

            // raise event
            if (this.AfterClear != null)
            {
                // invoke each attached delegate  
                foreach (var handler in (EventHandler[])this.AfterClear.GetInvocationList())
                {
                    // setup args
                    this.args.Cancel = false;

                    // raise event
                    handler(this, this.args);
                }
            }
        }

        ///<summary>
        /// Gets the number of elements contained in the dictionary. />.
        ///</summary>
        ///<returns>
        /// The number of elements contained in the dictionary />.
        ///</returns>
        public virtual int Count
        {
            get { return this.dictionaryItems.Count; }
        }


        ///<summary>
        ///                    Removes the first occurrence of a specific object from the <see cref="T:System.Collections.Generic.ICollection`1" />.
        ///</summary>
        ///<returns>
        ///true if <paramref name="pair" /> was successfully removed from the dictionary otherwise, false. This method also returns false if 
        /// pair is not found in the dictionary.
        ///</returns>
        ///<param name="pair">The object to remove from the dictionary. </param>
        ///<exception cref="T:System.NotSupportedException">
        ///  The dictionary is read-only.
        ///  </exception>
        public virtual bool Remove(KeyValuePair<TKey, TValue> pair)
        {
            return this.Remove(pair.Key);
        }

        ///<summary>
        /// Removes the element with the specified key from the <see cref="T:System.Collections.Generic.IDictionary`2" />.
        ///</summary>
        ///<returns>
        ///true if the element is successfully removed; otherwise, false.  This method also returns false if <paramref name="Key" /> was not found in the original <see cref="T:System.Collections.Generic.IDictionary`2" />.
        ///</returns>
        ///<param name="Key">The key of the element to remove.</param>
        ///<exception cref="T:System.ArgumentNullException">Key is null.</exception>
        ///<exception cref="T:System.NotSupportedException">
        /// The dictionary is read-only.
        /// </exception>
        public virtual bool Remove(TKey Key)
        {
            if (!this.DoBeforeEvent(this.BeforeRemove, Key, default(TValue))) return false;

            // remove item
            this.dictionaryItems.Remove(Key);

            this.DoAfterEvent(this.AfterRemove, default(TValue), Key);

            return true;
        }

        ///<summary>
        /// Gets or sets the element with the specified key.
        ///</summary>
        ///<returns>
        /// The element with the specified key.
        ///</returns>
        ///<param name="Key">The key of the element to get or set.</param>
        ///<exception cref="T:System.ArgumentNullException">Key is null.</exception>
        ///<exception cref="T:System.Collections.Generic.KeyNotFoundException">
        /// The property is retrieved and Key is not found.
        /// </exception>
        ///<exception cref="T:System.NotSupportedException">
        /// The property is set and the dictionary is read-only.
        /// </exception>
        public virtual TValue this[TKey Key]
        {
            get { return this.dictionaryItems[Key]; }
            set
            {
                if (this.DoBeforeEvent(this.BeforeChange, Key, default(TValue))) return;

                // change item
                this.dictionaryItems[Key] = value;

                this.DoAfterEvent(this.AfterChange, default(TValue), Key);
            }
        }


        #region " IDisposable Support "
        private bool disposedValue; // To detect redundant calls

        // IDisposable
        protected virtual void Dispose(bool disposing)
        {
            if (!this.disposedValue)
            {
                if (disposing)
                {
                    // TODO: free unmanaged resources when explicitly called
                    this.dictionaryItems = null;
                }

                // TODO: free shared unmanaged resources
            }
            this.disposedValue = true;
        }

        // This code added by Visual Basic to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion

        ///<summary>
        ///  Determines whether the dictionary contains an element with the specified key.
        ///</summary>
        ///<returns>
        ///true if the dictionary contains an element with the key; otherwise, false.
        ///</returns>
        ///<param name="key">The key to locate in the dictionary.</param>
        ///<exception cref="T:System.ArgumentNullException">Key is null.</exception>
        public virtual bool ContainsKey(TKey key)
        {
            return this.dictionaryItems.ContainsKey(key);
        }

        ///<summary>
        ///  Gets the value associated with the specified key.
        ///</summary>
        ///<returns>
        ///true if the object that implements <see cref="T:System.Collections.Generic.IDictionary`2" /> contains an element with the specified key; otherwise, false.
        ///</returns>
        ///<param name="key">The key whose value to get.</param>
        ///<param name="value">
        /// When this method returns, the value associated with the specified key, if the key is found; otherwise, the default value for the type 
        /// of the value parameter. This parameter is passed uninitialized.
        /// </param>
        ///<exception cref="T:System.ArgumentNullException">key is null.    </exception>
        public virtual bool TryGetValue(TKey key, out TValue value)
        {
            return this.dictionaryItems.TryGetValue(key, out value);
        }

        ///<summary>
        ///   Determines whether the <see cref="T:System.Collections.Generic.ICollection`1" /> contains a specific value.
        ///</summary>
        ///<returns>
        ///true if item is found in the dictionary otherwise, false.
        ///</returns>
        ///<param name="item">The object to locate in the dictionary.</param>
        public virtual bool Contains(KeyValuePair<TKey, TValue> item)
        {
            return this.dictionaryItems.ContainsKey(item.Key) || this.dictionaryItems.ContainsValue(item.Value);
        }

        ///<summary>
        ///  Copies the elements of the dictionary to an array , starting at a particulararray index.
        ///</summary>
        ///<param name="array">The one-dimensional array that is the destination of the elements copied from dictionary. The array must have zero-based indexing.</param>
        ///<param name="arrayIndex">The zero-based index in array at which copying begins.</param>
        ///<exception cref="T:System.ArgumentNullException">array is null.</exception>
        ///<exception cref="T:System.ArgumentOutOfRangeException">arrayIndex is less than 0.</exception>
        ///<exception cref="T:System.ArgumentException">array  is multidimensional.
        ///                
        ///                    -or-
        ///                    arrayIndex is equal to or greater than the length of array.
        ///                
        ///                    -or-
        ///                
        ///                    The number of elements in the source dictionary is greater than the available space from arrayIndex to the end of the destination array.
        ///                
        ///                    -or-
        ///                
        ///                    Cannot be cast automatically to the type of the destination array.
        ///                </exception>
        public virtual void CopyTo(KeyValuePair<TKey, TValue>[] array, int arrayIndex)
        {
            if (array == null) throw new ArgumentNullException("array");
            if ((arrayIndex < 0) || (arrayIndex > array.Length)) throw new ArgumentOutOfRangeException("arrayIndex");
            if ((array.Length - arrayIndex) < this.Count) throw new ArgumentException("array is not large engough to hold the items from this dictionary.");

            foreach (var pair in this.dictionaryItems)
            {
                array[arrayIndex++] = new KeyValuePair<TKey, TValue>(pair.Key, pair.Value);
            }
        }

        ///<summary>
        /// Gets a array of items stored in the dictionary.
        ///</summary>
        ///<returns>Returns a array of items stored in the dictionary.</returns>
        public virtual KeyValuePair<TKey, TValue>[] ToArray()
        {
            KeyValuePair<TKey, TValue>[] entries = new KeyValuePair<TKey, TValue>[this.dictionaryItems.Count];

            int index = 0;
            foreach (var pair in this.dictionaryItems)
            {
                entries[index++] = new KeyValuePair<TKey, TValue>(pair.Key, pair.Value);
            }
            return entries;
        }

        ///<summary>
        /// Gets a array of keys stored in the dictionary.
        ///</summary>
        ///<returns>Returns a array of keys stored in the dictionary.</returns>
        public virtual TKey[] KeysToArray()
        {
            TKey[] entries = new TKey[this.dictionaryItems.Count];

            int index = 0;
            foreach (var pair in this.dictionaryItems)
            {
                entries[index++] = pair.Key;
            }
            return entries;
        }

        ///<summary>
        /// Gets a array of values stored in the dictionary.
        ///</summary>
        ///<returns>Returns a array of values stored in the dictionary.</returns>
        public virtual TValue[] ValuesToArray()
        {
            TValue[] entries = new TValue[this.dictionaryItems.Count];

            int index = 0;
            foreach (var pair in this.dictionaryItems)
            {
                entries[index++] = pair.Value;
            }
            return entries;
        }
        
        ///<summary>
        ///  Gets a value indicating whether the dictionary is read-only.
        ///</summary>
        ///<returns>
        ///true if the dictionary is read-only; otherwise, false.
        ///</returns>
        public bool IsReadOnly
        {
            get { return false; }
        }

        ///<summary>
        ///  Returns an enumerator that iterates through the dictionary.
        ///</summary>
        ///<returns>
        ///  A IEnumerator that can be used to iterate through the dictionary.
        ///</returns>
        public IEnumerator<KeyValuePair<TKey, TValue>> GetEnumerator()
        {
            return this.dictionaryItems.GetEnumerator();
        }

        ///<summary>
        ///  Returns an enumerator that iterates through the dictionary.
        ///</summary>
        ///<returns>
        ///  A IEnumerator that can be used to iterate through the dictionary.
        ///</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.dictionaryItems.GetEnumerator();
        }

        ///<summary>
        /// Gets an ICollection containing the keys of the dictionary.
        ///</summary>
        ///<returns>
        /// An ICollection containing the keys of the dictionary.
        ///</returns>
        ICollection<TKey> IDictionary<TKey, TValue>.Keys
        {
            get { return this.dictionaryItems.Keys; }
        }

        ICollection<TValue> IDictionary<TKey, TValue>.Values
        {
            get { return this.dictionaryItems.Values; }
        }
    }

}
